;;;; SPDX-FileCopyrightText: Atlas Engineer LLC
;;;; SPDX-License-Identifier: BSD-3-Clause

(in-package :prompter)

(define-class sync-queue ()
  ((ready-sources '()
                  :type list
                  :export nil
                  :documentation
                  "List of ready sources.")
   (ready-channel (make-channel nil)
                  :type calispel:channel
                  :export nil
                  :documentation
                  "Communication channel with the `update' thread."))
  (:accessor-name-transformer (hu.dwim.defclass-star:make-name-transformer name))
  (:documentation "This object is used to memorize which sources are ready for a
given input.
A new object is created on every new input."))

(defvar *default-history-size* 1000)    ; TODO: Export?

(declaim (ftype (function (&key (:size fixnum)) containers:ring-buffer-reverse) make-history))
(defun make-history (&key (size *default-history-size*))
  "Return a new ring buffer."
  (the (values cl-containers:ring-buffer-reverse &optional)
       (containers:make-ring-buffer size :last-in-first-out)))

;; Same as `source' as to why we wrap in `eval-always'.
(sera:eval-always
  (define-class prompter ()
    ((input ""
            :accessor nil
            :reader input
            :documentation
            "User input.")

     (prompt ""
             :documentation
             "Prefix to the user input.")

     (sources '()
              :type (or null source (cons source))
              :documentation "List of `source's.
For convenience, if the initarg is a single source (that is, not inside a list),
it is automatically wrapped into a list upon initialization.
If the source is designated by a symbol, then it is automatically instantiated
with `make-instance'.")

     (selection '()
                ;; TODO: Index by (source-index suggestion-index) instead?
                ;; TODO: Use structure?
                :type list
                :export nil
                :reader selection
                :documentation "A pair of source and suggestion index.")

     (constructor nil
                  :type (or null function)
                  :documentation
                  "Function called with the prompter as argument.")

     (before-destructor nil
                        :type (or null function)
                        :documentation
                        "First function called with no parameters when calling the
`destroy' function over this prompter.
It's called before the sources are cleaned up.")
     (after-destructor nil
                       :type (or null function)
                       :documentation
                       "Last function called with no parameters when calling the
`destroy' function over this prompter.
It's called after the sources are cleaned up.

Note that the function is executed *before* performing an action.")

     (history (make-history)
              :type (or containers:ring-buffer-reverse null)
              :documentation
              "History of inputs for the prompter.
If nil, no history is used.")

     (keymap nil
             :type (or null keymap:keymap)
             :documentation
             "Keymap for the prompter.
Useful, say, to switch source.
It takes precedence over individual source keymaps.")

     (help-message ""                   ; TODO: Use.
                   :type (or string function)
                   :documentation
                   "Help message for this prompter.
It can be a function of one argument, the prompter, which returns a string.")
     (result-channel (make-channel 1)
                     :type calispel:channel
                     :documentation
                     "Channel to which the selection is sent on exit.
Caller should also listen to `interrupt-channel' to know if the minibuffer is quitted.")
     (interrupt-channel (make-channel 1)
                        :type calispel:channel
                        :documentation
                        "Channel to which an arbitrary value is written on exit.
See also `result-channel'.")

     (sync-queue nil
                 :type (or null sync-queue)
                 :export nil
                 :documentation "See `sync-queue' class documentation.")

     (returned-p nil
                 :type boolean
                 :documentation
                 "If non-nil, prompter has returned.
This is useful to know if prompter was cancelled or not."))
    (:export-class-name-p t)
    (:export-accessor-names-p t)
    (:accessor-name-transformer (hu.dwim.defclass-star:make-name-transformer name))
    (:documentation "The prompter is an interface for user interactions.
A prompter object holds multiple sources (of type `source') which
contain a list of `suggestion's.

You can call `destroy' to call the registered termination functions of the
prompter and its sources.

Suggestions are computed asynchronously when `input' is updated.
Use `all-ready-p' and `next-ready-p' to know when the prompter is ready.
Sources suggestions can be retrieved, possibly partially, even when the
compution is not finished.")))

(defun update-sources (prompter &optional (text ""))
  (setf (sync-queue prompter) (make-instance 'sync-queue))
  (mapc (lambda (source) (update source text (ready-channel (sync-queue prompter))))
        (sources prompter)))

(defmethod initialize-instance :after ((prompter prompter) &key)
  (unless (stringp (prompt prompter))
    (setf (prompt prompter) (write-to-string (prompt prompter))))
  (unless (stringp (input prompter))
    (setf (input prompter) (write-to-string (input prompter))))
  (flet ((ensure-sources (specifiers)
           (mapcar (lambda (source-specifier)
                     (cond
                       ((source-p source-specifier)
                        source-specifier)
                       ((and (symbolp source-specifier)
                             (c2cl:subclassp source-specifier 'source))
                        (make-instance source-specifier))
                       (t (error "Bad source specifier ~s." source-specifier))))
                   (uiop:ensure-list specifiers))))
    (setf (sources prompter) (ensure-sources (sources prompter))))
  (setf (selection prompter) (list (first (sources prompter)) 0))
  (maybe-funcall (constructor prompter) prompter)
  (update-sources prompter)
  prompter)

(defmethod (setf selection) (value (prompter prompter))
  (setf (slot-value prompter 'selection) value)
  (let ((source (selected-source prompter)))
    (when (follow-p source)
      (if (< 0 (follow-delay source))
          (bt:make-thread
           (lambda ()
             (sleep (follow-delay source))
             (call-persistent-action prompter)))
          (call-persistent-action prompter)))))

(export-always 'input)
(defmethod (setf input) (text (prompter prompter))
  "Update PROMPTER sources and return TEXT."
  (let ((old-input (slot-value prompter 'input)))
    (unless (string= old-input text)
      (setf (slot-value prompter 'input) text)
      (update-sources prompter text)
      ;; TODO: Update `selection' when `update' is done.
      (setf (selection prompter) (list (first (sources prompter)) 0))))
  text)

(export-always 'destroy)
(defun destroy (prompter)
  "First call `before-destructor', then call all the source destructors, finally call
`after-destructor'.
Signal destruction by sending a value to PROMPTER's `interrupt-channel'."
  (maybe-funcall (before-destructor prompter))
  (mapc (lambda (source) (maybe-funcall (destructor source) prompter source))
        (sources prompter))
  (maybe-funcall (after-destructor prompter))
  ;; TODO: Interrupt before or after destructor?
  (calispel:! (interrupt-channel prompter) t))

(export-always 'call-persistent-action)
(defun call-persistent-action (prompter)
  (sera:and-let* ((action (persistent-action (selected-source prompter)))
                  (suggestion (selected-suggestion prompter)))
    (funcall action
             (value suggestion))))

(defun select (prompter steps &key wrap-over-p)
  "Select suggestion by jumping STEPS forward.
If STEPS is 0, do nothing.
If STEPS is negative, go backward.
If the currently selected suggestion is the last one of the current source, go
to next source, or previous source if STEPS is negative."
  (unless (= 0 steps)
    (labels ((index->source (index &optional (sources (sources prompter)))
               (let ((limit (length (suggestions (first sources)))))
                 (if (or (< index limit)
                         ;; Can happen when INDEX is beyond the total length.
                         (null (rest sources)))
                     (first sources)
                     (index->source (- index limit) (rest sources)))))
             (source-length (sources)
               (reduce #'+ (mapcar (lambda (source) (length (suggestions source)))
                                   sources)))
             (previous-sources (source)
               (let ((current-source-position (position source (sources prompter))))
                 (subseq (sources prompter) 0 current-source-position))))
      (let* ((limit (source-length (sources prompter)))
             (previous-sources (previous-sources (first (selection prompter))))
             (index (+ (second (selection prompter))
                       (source-length previous-sources)))
             (new-index (+ index steps)))
        (setf new-index
              (if wrap-over-p
                  (mod new-index limit)
                  (alex:clamp new-index 0 (1- limit))))
        (let* ((new-source (index->source new-index))
               (relative-index (- new-index
                                  (source-length (previous-sources new-source)))))
          (setf (selection prompter)
                (list new-source relative-index)))))))

(export-always 'select-next)
(defun select-next (prompter &optional (steps 1))
  "Select element by jumping STEPS forward.
If STEPS is 0, do nothing.
If STEPS is negative, go backward."
  (select prompter steps))

(export-always 'select-previous)
(defun select-previous (prompter &optional (steps 1))
  "Select element by jumping STEPS forward.
If STEPS is 0, do nothing.
If STEPS is negative, go forward."
  (select prompter (- steps)))

(export-always 'select-next-source)
(defun select-next-source (prompter &optional (steps 1))
  "Jumping STEPS source forward and select first suggestion.
If STEPS is 0, do nothing.
If STEPS is negative, go backward and select last suggestion."
  (unless (= 0 steps)
    (let* ((source-index (position (selected-source prompter)
                                   (sources prompter)))
           (new-source (nth (alex:clamp (+ steps source-index) 0 (1- (length (sources prompter))))
                            (sources prompter)))
           (suggestion-index (if (< 0 steps)
                                 0
                                 (1- (length (suggestions new-source))))))
      (setf (selection prompter)
            (list new-source suggestion-index)))))

(export-always 'select-previous-source)
(defun select-previous-source (prompter &optional (steps 1))
  "Jumping STEPS source backward and select last suggestion.
If STEPS is 0, do nothing.
If STEPS is negative, go forward and selection first suggestion."
  (unless (= 0 steps)
    (select-next-source prompter (- steps))))

(export-always 'select-first)
(defun select-first (prompter)
  "Select first element."
  (setf (selection prompter)
        (list (first (sources prompter)) 0)))

(export-always 'select-last)
(defun select-last (prompter)
  "Select last element."
  (let ((last-source (first (last (sources prompter)))))
    (setf (selection prompter)
          (list last-source
                (1- (length (suggestions last-source)))))))

(export-always 'toggle-mark)
(defun toggle-mark (prompter)
  (when (multi-selection-p (selected-source prompter))
    (multiple-value-bind (suggestion source)
        (selected-suggestion prompter)
      (with-accessors ((marks marks)) source
        (let ((value (value suggestion)))
          (if (find value marks)
              (setf marks (delete value marks))
              (push value marks)))))))

(export-always 'mark-all)
(defun mark-all (prompter)
  (let ((source (selected-source prompter)))
    (when (multi-selection-p source)
      (alex:unionf (marks source)
                   (mapcar #'value (suggestions source))))))

(export-always 'unmark-all)
(defun unmark-all (prompter)
  (let ((source (selected-source prompter)))
    (when (multi-selection-p source)
      (with-accessors ((marks marks)
                       (suggestions suggestions))
          source
        (setf marks
              (set-difference marks
                              (mapcar #'value suggestions)))))))

(export-always 'toggle-mark-all)
(defun toggle-mark-all (prompter)
  (let ((source (selected-source prompter)))
    (when (multi-selection-p source)
      (with-accessors ((suggestions suggestions)
                       (marks marks))
          source
        (let ((suggestion-values (mapcar #'value suggestions)))
          (setf marks
                (cond
                  ((subsetp marks suggestion-values)
                   (set-difference suggestion-values marks))
                  ((subsetp suggestion-values marks)
                   (set-difference marks suggestion-values))
                  (t ; When the intersection of suggestion-values and marks is non-trivial.
                   (set-difference
                    (union marks suggestion-values)
                    (intersection marks suggestion-values))))))))))

(defun resolve-selection (prompter)     ; TODO: Write tests for this!
  "Return the list of selected values.
If there is no marks, the current selection value is returned as a list of one element.
For instance, if the selected element value is NIL, this returns '(NIL).
If there is no element, NIL is returned."
  (or (all-marks prompter)
      (mapcar #'value (uiop:ensure-list (selected-suggestion prompter)))
      (and (not (must-match-p (selected-source prompter))) ; TODO: Remove when we remove `must-match-p'.
           (slot-value prompter 'input))))

(export-always 'actions)
(defun actions (prompter)
  "Return the list of contextual actions.
Without marks, it's the list of actions for the current source.
With marks, it's the intersection of the actions of the sources that contain the
marked elements."
  (let ((marked-sources
          (remove-if (complement #'marks) (sources prompter))))
    (if marked-sources
        (reduce #'intersection (mapcar (lambda (source)
                                         (slot-value source 'actions))
                                       marked-sources))
        (slot-value (selected-source prompter) 'actions))))

(defun add-input-to-history (prompter)
  (unless (or (null (history prompter))
              (str:empty? (input prompter)))
    ;; TODO: Make sure we remove duplicates.
    (containers:insert-item (history prompter) (input prompter))))

(export-always 'return-selection)
(defun return-selection (prompter &optional (action (default-action prompter)))
  "Call action over selection and send the results to PROMPTER's `result-channel'.
The selection is the collection of marked suggestions across all sources.
If there is no marked suggestion, send the currently selected suggestion
instead."
  (unless action
    (setf action #'identity))
  (setf (returned-p prompter) t)
  (add-input-to-history prompter)
  (alex:if-let ((selection-value (resolve-selection prompter)))
    (let ((action-result (funcall action selection-value)))
      (calispel:! (result-channel prompter) action-result))
    (calispel:! (interrupt-channel prompter) t)))

(export-always 'return-input)
(defun return-input (prompter)
  "Send input to PROMPTER's `result-channel'."
  (setf (returned-p prompter) t)
  (add-input-to-history prompter)
  (calispel:! (result-channel prompter) (input prompter)))

(export-always 'toggle-follow)
(defun toggle-follow (prompter &optional (source (selected-source prompter)))
  "Toggle `follow-p' in SOURCE."
  (setf (follow-p source) (not (follow-p source))))

(export-always 'next-ready-p)
(defun next-ready-p (prompter &optional timeout)
  "Block and return next PROMPTER ready source.
It's the next source that's done updating.
If all sources are done, return t.
If timeout expires for one source, return nil."
  ;; We copy `sync-queue' here so that it remains the same object throughout
  ;; this function, since the slot is subject to be changed concurrently when
  ;; the input is edited.
  (let ((sync-queue (sync-queue prompter)))
    (if sync-queue
        (if (= (length (ready-sources sync-queue))
               (length (sources prompter)))
            t
            (multiple-value-bind (next-source ok)
                (calispel:? (ready-channel sync-queue) timeout)
              (cond
                ((not ok)
                 nil)
                ((null next-source)
                 nil)
                (t
                 (push next-source (ready-sources sync-queue))
                 next-source))))
        ;; No sync-queue if no input was ever set.
        t)))

(export-always 'all-ready-p)
(defun all-ready-p (prompter &optional timeout)
  "Return non-nil when all prompter sources are ready.
After timeout has elapsed for one source, return nil."
  (sera:nlet check ((next-source (next-ready-p prompter timeout)))
    (cond
      ((eq t next-source)
       t)
      ((null next-source)
       nil)
      (t
       (check (next-ready-p prompter timeout))))))

(export-always 'make)
(define-function make
    (append '(&rest args)
            `(&key ,@(public-initargs 'prompter)))
  "Return prompter object."
  (apply #'make-instance 'prompter args))

(export-always 'selected-source)
(defun selected-source (prompter)
  (first (selection prompter)))

(export-always 'selected-suggestion)
(defun selected-suggestion (prompter)
  "Return selected prompt-buffer suggestion.
Return source as second value."
  (let* ((source (first (selection prompter))))
    (values (nth (second (selection prompter)) (suggestions source)) source)))

(export-always 'selected-suggestion-position)
(defun selected-suggestion-position (prompter)
  "Return selected prompt-buffer suggestion position among current source
suggestions."
  (second (selection prompter)))

(export-always 'all-marks)
(defun all-marks (prompter)
  "Return the list of the marked suggestion values in the prompter."
  (alex:mappend #'marks (sources prompter)))

(defun default-action (prompter)
  (first (actions prompter)))

(export-always 'resume)
(defun resume (prompter)
  "Calls each source `resumer' function over the source.
This is meant to be called when a prompter is resumed."
  (mapc (lambda (source)
          (maybe-funcall (resumer source) source))
        (sources prompter)))
